////
// 🦠 Corona-Warn-App
//

import Foundation

protocol TraceWarningMatching {

	func matchAndStore(package: SAPDownloadedPackage)
	func calculateOverlap(checkin: Checkin, warning: SAP_Internal_Pt_TraceTimeIntervalWarning) -> Int
	func calculateOverlap(checkin: Checkin, match: TraceTimeIntervalMatch) -> Int

}

final class TraceWarningMatcher: TraceWarningMatching {

	// MARK: - Init

	init(
		eventStore: EventStoringProviding
	) {
		self.eventStore = eventStore
	}

	// MARK: - Overrides

	// MARK: - Protocol TraceWarningMatching

	func matchAndStore(package: SAPDownloadedPackage) {
		Log.info("[TraceWarningMatching] Start matching TraceTimeIntervalWarnings against Checkins. ", log: .checkin)

		guard let warningPackage = try? SAP_Internal_Pt_TraceWarningPackage(serializedData: package.bin) else {
			Log.error("[TraceWarningMatching] Failed to decode SAPDownloadedPackage", log: .checkin)
			return
		}
		matchAndStore(package: warningPackage)
	}

	func calculateOverlap(checkin: Checkin, warning: SAP_Internal_Pt_TraceTimeIntervalWarning) -> Int {
		let endIntervalNumber = warning.startIntervalNumber + warning.period

		return calculateOverlap(
			checkin: checkin,
			startIntervalNumber: Int(warning.startIntervalNumber),
			endIntervalNumber: Int(endIntervalNumber)
		)
	}

	func calculateOverlap(checkin: Checkin, match: TraceTimeIntervalMatch) -> Int {
		calculateOverlap(
			checkin: checkin,
			startIntervalNumber: match.startIntervalNumber,
			endIntervalNumber: match.endIntervalNumber
		)
	}

	// MARK: - Internal

	func matchAndStore(package: SAP_Internal_Pt_TraceWarningPackage) {
		Log.info("[TraceWarningMatching] Package interval number: \(package.intervalNumber)", log: .checkin)

		for warning in package.timeIntervalWarnings {
			Log.info("[TraceWarningMatching] Warning startIntervalNumber: \(warning.startIntervalNumber), period: \(warning.period)", log: .checkin)
			Log.debug("[TraceWarningMatching] Warning : \(warning)", log: .checkin)

			// Filter checkins with same id hash.
			var checkins: [Checkin] = eventStore.checkinsPublisher.value.filter {
				$0.traceLocationIdHash == warning.locationIDHash
			}

			guard !checkins.isEmpty else {
				continue
			}
			Log.info("[TraceWarningMatching] Found \(checkins.count) matches with the same location id hash.", log: .checkin)

			// Filter checkins where the warning overlaps the timeframe.
			checkins = checkins.filter {
				calculateOverlap(checkin: $0, warning: warning) > 0
			}

			guard !checkins.isEmpty else {
				continue
			}
			Log.info("[TraceWarningMatching] Found \(checkins.count) overlaping matches.", log: .checkin)

			// Persist checkins and warning as matches.
			for checkin in checkins {
				let match = TraceTimeIntervalMatch(
					id: 0, // createTraceTimeIntervalMatch will ignore this id. The id is generated by the database.
					checkinId: checkin.id,
					traceWarningPackageId: Int(package.intervalNumber),
					traceLocationId: checkin.traceLocationId,
					transmissionRiskLevel: Int(warning.transmissionRiskLevel),
					startIntervalNumber: Int(warning.startIntervalNumber),
					endIntervalNumber: Int(warning.startIntervalNumber + warning.period)
				)

				Log.info("[TraceWarningMatching] Persist match with checkin.id: \(checkin.id) and package.intervalNumber: \(package.intervalNumber). ", log: .checkin)
				eventStore.createTraceTimeIntervalMatch(match)
			}
		}
	}

	// MARK: - Private
	
	private let eventStore: EventStoringProviding

	// Algorithm from: https://github.com/corona-warn-app/cwa-app-tech-spec/blob/proposal/event-registration-mvp/sample-code/presence-tracing/pt-calculate-overlap.js

	private func calculateOverlap(checkin: Checkin, startIntervalNumber: Int, endIntervalNumber: Int) -> Int {
		func toTimeInterval(_ intervalNumber: Int) -> TimeInterval {
			TimeInterval(intervalNumber * 600)
		}

		let warningStartTimestamp = toTimeInterval(startIntervalNumber)
		let warningEndTimestamp = toTimeInterval(endIntervalNumber)

		let overlapStartTimestamp = max(checkin.checkinStartDate.timeIntervalSince1970, warningStartTimestamp)
		let overlapEndTimestamp = min(checkin.checkinEndDate.timeIntervalSince1970, warningEndTimestamp)
		let overlapInSeconds = overlapEndTimestamp - overlapStartTimestamp

		if overlapInSeconds < 0 {
			return 0
		} else {
			return Int(round(overlapInSeconds / 60))
		}
	}

}
